// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "../Asset.hpp"
#include "asset/AssetLoader.hpp"
#include "util/flag.hpp"
#include "display/font/truetype/TrueTypeFaceTexture.hpp"
#include "util/UsageLock.hpp"

namespace Dawn {
  enum TrueTypeAssetState {
    TRUE_TYPE_ASSET_STATE_INITIAL,
    TRUE_TYPE_ASSET_STATE_OPEN,
    TRUE_TYPE_ASSET_STATE_VALIDATE_HEADER,
    TRUE_TYPE_ASSET_STATE_VALIDATE_VERSION,
    TRUE_TYPE_ASSET_STATE_READ_VARIANT_COUNT,
    TRUE_TYPE_ASSET_STATE_READ_VARIANT,
    TRUE_TYPE_ASSET_STATE_ADJUSTING_OFFSETS,
    TRUE_TYPE_ASSET_STATE_INIT_FREETYPE,
    TRUE_TYPE_ASSET_STATE_READY
  };

  struct TrueTypeAssetStyle {
    flag_t style;
    size_t dataSize;
    size_t dataOffset;
  };

  class TrueTypeAsset : public Asset {
    protected:
      UsageLock locks;
      AssetLoader loader;
      FT_Library fontLibrary;
      enum TrueTypeAssetState state = TRUE_TYPE_ASSET_STATE_INITIAL;
      std::vector<struct TrueTypeAssetStyle> assetStyles;
      std::vector<TrueTypeFaceTexture*> textures;
      std::map<usagelockid_t, TrueTypeFaceTexture*> textureByLock;
      std::map<struct TrueTypeFaceTextureStyle, TrueTypeFaceTexture*>
        textureByStyle
      ;
      std::map<TrueTypeFaceTexture*, std::vector<usagelockid_t>> locksByTexture;

    public:
      TrueTypeAsset(const std::string name);
      void updateSync() override;
      void updateAsync() override;

      /**
       * Create a lock for a specific style. Locks will ensure that the font is
       * held loaded until it is no longer required.
       * 
       * @param style Style to lock.
       * @return A unique lock ID for this style.
       */
      usagelockid_t lock(struct TrueTypeFaceTextureStyle style);

      /**
       * Get a texture by a previous lock ID.
       * 
       * @param lock Lock to get the texture of.
       * @return Matching texture by this ID.
       */
      TrueTypeFaceTexture * getTexture(usagelockid_t lock);

      /**
       * Releases a previously held font lock.
       * 
       * @param lock Lock to release/unlock.
       */
      void unlock(usagelockid_t lock);

      ~TrueTypeAsset();
  };
}
