// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UIComponent.hpp"

using namespace Dawn;

UIComponent::UIComponent(SceneItem *item) :
  SceneItemComponent(item),
  alignment(glm::vec4(0, 0, 128, 128)),
  alignUnitLeft(UI_COMPONENT_ALIGN_UNIT_SCALE),
  alignUnitTop(UI_COMPONENT_ALIGN_UNIT_SCALE),
  alignUnitRight(UI_COMPONENT_ALIGN_UNIT_SCALE),
  alignUnitBottom(UI_COMPONENT_ALIGN_UNIT_SCALE),
  alignX(UI_COMPONENT_ALIGN_START),
  alignY(UI_COMPONENT_ALIGN_START),
  alignmentNeedsUpdating(true)
{
  
}

UIComponentDimensional * UIComponent::getParentDimensional() {
  auto parent = this->transform->getParent();
  if(parent == nullptr) return nullptr;
  auto dimensional = parent->item->getComponent<UIComponentDimensional>();
  assertNotNull(dimensional, "UIComponent::getParentDimensional: Parent must have a UIComponentDimensional");
  return dimensional;
}

void UIComponent::updateAlignment() {
  if(!this->alignmentNeedsUpdating) return;
  
  auto align = (glm::vec4)this->alignment;
  auto dimensional = this->getParentDimensional();
  auto translate = this->transform->getLocalPosition();

  assertNotNull(dimensional, "UIComponent::updateAlignment: Parent must have a UIComponentDimensional");

  parentInnerWidth = dimensional->getContentWidth();
  parentInnerHeight = dimensional->getContentHeight();

  // Should we be doing width first, or height first?
  bool_t heightFirst = (
    this->alignUnitLeft == UI_COMPONENT_ALIGN_UNIT_RATIO ||
    this->alignUnitRight == UI_COMPONENT_ALIGN_UNIT_RATIO
  );

  if(heightFirst) {
    UIComponent::calculateDimensions(
      this->alignY,
      this->alignUnitTop,
      this->alignUnitBottom,
      &translate.y,
      &this->height,
      parentInnerHeight,
      this->getContentHeight(),
      this->width,
      glm::vec2(align[1], align[3])
    );
    UIComponent::calculateDimensions(
      this->alignX,
      this->alignUnitLeft,
      this->alignUnitRight,
      &translate.x,
      &this->width,
      parentInnerWidth,
      this->getContentWidth(),
      this->height,
      glm::vec2(align[0], align[2])
    );
  } else {
    UIComponent::calculateDimensions(
      this->alignX,
      this->alignUnitLeft,
      this->alignUnitRight,
      &translate.x,
      &this->width,
      parentInnerWidth,
      this->getContentWidth(),
      this->height,
      glm::vec2(align[0], align[2])
    );
    UIComponent::calculateDimensions(
      this->alignY,
      this->alignUnitTop,
      this->alignUnitBottom,
      &translate.y,
      &this->height,
      parentInnerHeight,
      this->getContentHeight(),
      this->width,
      glm::vec2(align[1], align[3])
    );
  }


  translate.x += dimensional->getChildOffsetX();
  translate.y += dimensional->getChildOffsetY();

  this->transform->setLocalPosition(translate);
  this->alignmentNeedsUpdating = false;
  this->eventAlignmentUpdated.invoke();
}

float_t UIComponent::getWidthFromAlignment() {
  switch(this->alignX) {
    case UI_COMPONENT_ALIGN_STRETCH:
    case UI_COMPONENT_ALIGN_START:
    case UI_COMPONENT_ALIGN_MIDDLE:
      return alignment._realValue[2];

    case UI_COMPONENT_ALIGN_END:
      return alignment._realValue[0];

    default:
      assertUnreachable("UIComponent::getWidthFromAlignment: Unknown alignment");
      return -1;
  }
}

float_t UIComponent::getHeightFromAlignment() {
  switch(this->alignY) {
    case UI_COMPONENT_ALIGN_STRETCH:
    case UI_COMPONENT_ALIGN_START:
    case UI_COMPONENT_ALIGN_MIDDLE:
      return alignment._realValue[3];

    case UI_COMPONENT_ALIGN_END:
      return alignment._realValue[1];

    default:
      assertUnreachable("UIComponent::getWidthFromAlignment: Unknown alignment");
      return -1;
  }
}

float_t UIComponent::calculateAlignmentValue(
  float_t alignmentValue,
  float_t parentSize,
  float_t ratioSize,
  enum UIComponentAlignUnit unit
) {
  if(unit == UI_COMPONENT_ALIGN_UNIT_SCALE) return alignmentValue;
  if(unit == UI_COMPONENT_ALIGN_UNIT_PERCENT) return (alignmentValue / 100.0f) * parentSize;
  if(unit == UI_COMPONENT_ALIGN_UNIT_RATIO) return (alignmentValue / 100.0f) * ratioSize;
  assertUnreachable("UIComponent::calculateAlignmentValue: Unknown alignment unit");
  return -1;
}

void UIComponent::calculateDimensions(
  enum UIComponentAlign align,
  enum UIComponentAlignUnit unitStart,
  enum UIComponentAlignUnit unitEnd,
  float_t *position,
  float_t *size,
  float_t outerSize,
  float_t innerSize,
  float_t ratioSize,
  glm::vec2 alignment
) {
  assertNotNull(position, "UIComponent::calculateDimensions: Position cannot be null");
  assertNotNull(size, "UIComponent::calculateDimensions: Size cannot be null");

  switch(align) {
    case UI_COMPONENT_ALIGN_STRETCH: {
      // For stretch:
      // Alignment 0 becomes START offset, Alignment 1 becomes END Offset
      *position = UIComponent::calculateAlignmentValue(
        alignment[0],
        outerSize,
        ratioSize,
        unitStart
      );
      *size = outerSize - (*position + UIComponent::calculateAlignmentValue(
        alignment[1],
        outerSize,
        ratioSize,
        unitEnd
      ));
      break;
    }

    case UI_COMPONENT_ALIGN_START: {
      // For start:
      // Alignment 0 becomes START offset, Alignment 1 becomes SIZE
      *position = UIComponent::calculateAlignmentValue(
        alignment[0],
        outerSize,
        ratioSize,
        unitStart
      );
      *size = UIComponent::calculateAlignmentValue(
        alignment[1],
        outerSize,
        ratioSize,
        unitEnd
      );
      break;
    }

    case UI_COMPONENT_ALIGN_MIDDLE: {
      *size =  UIComponent::calculateAlignmentValue(
        alignment[1],
        outerSize,
        ratioSize,
        unitEnd
      );
      *position = (outerSize / 2.0f) - ((*size) / 2.0f) + UIComponent::calculateAlignmentValue(
        alignment[0],
        outerSize,
        ratioSize,
        unitStart
      );
      break;
    }

    case UI_COMPONENT_ALIGN_END: {
      *size = UIComponent::calculateAlignmentValue(
        alignment[0],
        outerSize,
        ratioSize,
        unitStart
      );
      *position = outerSize - *size - UIComponent::calculateAlignmentValue(
        alignment[1],
        outerSize,
        ratioSize,
        unitEnd
      );
      break;
    }
    
    default:
      assertUnreachable("UIComponent::calculateDimensions: Unknown alignment");
      break;
  }
}

UICanvas * UIComponent::getCanvas() {
  // TODO: Cache this on first hit.
  auto parent = this->transform->getParent();
  while(parent != nullptr) {
    auto canvas = parent->item->getComponent<UICanvas>();
    if(canvas != nullptr) return canvas;
    parent = parent->getParent();
  }
  assertUnreachable("UIComponent::getCanvas: No canvas found");
  return nullptr;
}

float_t UIComponent::getWidth() {
  return this->width;
}

float_t UIComponent::getHeight() {
  return this->height;
}

float_t UIComponent::getChildOffsetX() {
  return 0;
}

float_t UIComponent::getChildOffsetY() {
  return 0;
}

void UIComponent::onStart() {
  useEffect([&]{
    this->alignmentNeedsUpdating = true;
  }, { &alignment, &alignX, &alignY });

  useEffect([&]{
    this->updateAlignment();

    auto child = this->item->findChildren<UIComponent>();
    auto itChild = child.begin();
    while(itChild != child.end()) {
      (*itChild)->alignmentNeedsUpdating = true;
      ++itChild;
    }
  }, alignmentNeedsUpdating)();
}