// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "EntityAIWalk.hpp"

using namespace Dawn;

EntityAIWalk::EntityAIWalk(SceneItem* item) : SceneItemComponent(item) {
  
}

std::vector<SceneItemComponent*> EntityAIWalk::getDependencies() {
  return {
    (this->entityMove = item->getComponent<EntityMove>())
  };
}

void EntityAIWalk::onStart() {
  assertNotNull(this->entityMove);

  useEvent([&](float_t delta) {
    switch(this->mode) {
      case ENTITY_AI_WALK_MODE_FOLLOW_TARGET:
        this->updateFollowTarget(delta);
        break;

      case ENTITY_AI_WALK_MODE_WANDER:
        this->updateWander(delta);
        break;

      case ENTITY_AI_WALK_MODE_STAY:
        this->entityMove->direction = glm::vec2(0, 0);
        break;
    }
  }, getScene()->eventSceneUpdate);
}

void EntityAIWalk::updateWander(float_t delta) {
  // If we are stunned, give up wandering
  auto health = item->getComponent<EntityHealth>();
  if(health != nullptr && health->isStunned()) {
    wanderTimeToNextDecision = 0;
    return;
  }

  // Do we need to make a new decision?
  if(wanderTimeToNextDecision <= 0) {
    wanderTimeToNextDecision = randRange(wanderTimeRandomRange.x, wanderTimeRandomRange.y);

    // Do we want to move?
    float_t n = randRange(0.0f, 1.0f);
    if(n < wanderDoNotMoveChance) {
      wanderDestination = physics3Dto2D(transform->getLocalPosition());
      return;
    }

    wanderDestination = glm::vec2(
      randRange(-wanderDistanceRandomRange.x, wanderDistanceRandomRange.x),
      randRange(-wanderDistanceRandomRange.y, wanderDistanceRandomRange.y)
    );

    if(wanderSingleAxisOnly) {
      if(mathAbs<float_t>(wanderDestination.x) > mathAbs<float_t>(wanderDestination.y)) {
        wanderDestination.x = mathRound<float_t>(wanderDestination.x);
        wanderDestination.y = 0;
      } else {
        wanderDestination.x = 0;
        wanderDestination.y = mathRound<float_t>(wanderDestination.y);
      }
    }

    wanderDestination += physics3Dto2D(transform->getLocalPosition());
  } else {
    wanderTimeToNextDecision -= delta;
  }

  // Get the direction to move in
  glm::vec2 diff = wanderDestination - physics3Dto2D(transform->getLocalPosition());

  // Stop moving if we're close enough
  if(glm::length(diff) < 0.1f) {
    entityMove->direction = glm::vec2(0, 0);
  } else {
    entityMove->direction = diff;
  }
}

void EntityAIWalk::updateFollowTarget(float_t delta) {

}